@@ -1,5 +1,5 @@
-__author__ = 'Rens Holmer'
-__version__ = '0.0.19'
+__author__ = "Rens Holmer"
+__version__ = "0.0.22"
 
 from .tree import (  # noqa
     Tree,
@@ -11,9 +11,12 @@ from .sequence import (  # noqa
     alphabets,
     Sequence,
     SequenceReader,
+    BatchSequenceReader,
     AbstractSequenceCollection,
     SequenceCollection,
     MultipleSequenceAlignment,
     SequenceAnnotation,
     SequenceInterval,
 )
+
+from .ontology import Ontology, OntologyTerm  # noqa
